/* TokenTypes.java
 * Created on Nov 20, 2003
 * John Green
 *
 * Copyright (C) 2003-2011 Joanju Software (www.joanju.com)
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
package org.prorefactor.core;

import org.prorefactor.treeparserbase.JPTreeParserTokenTypes;

import com.joanju.proparse.NodeTypes;

/**
 * This static class provides information about Proparse's token types
 * as well as all of the static, final names for each token type.
 * NOTE: Now that Proparse has been rolled into this Java library, this
 * class is largely redundant with com.joanju.proparse.NodeTypes.
 * @See com.joanju.proparse.NodeTypes.
 */
public class TokenTypes implements JPTreeParserTokenTypes, TokenTypesI {

	private static TokenInfo [] allTokens = new TokenInfo[Last_Token_Number + 1];

	private static class TokenInfo {
		boolean isNatural = true;
		String keywordText = null;
	}

	static
	{
		for (int i = 1; i <= Last_Token_Number; i++) {
			String typeName = NodeTypes.getTypeName(i);
			if (typeName.length()==0)
				continue;
			TokenInfo info = new TokenInfo();
			info.isNatural = typeName.equals(typeName.toUpperCase());
			allTokens[i] = info;
		}
		assignDefaultText();
	}
	
	private static TokenInfo getInfo(int tokenType) {
		if (tokenType < 1 || tokenType > Last_Token_Number) return null;
		return allTokens[tokenType];
	}
	
	/** For integer token types that represent a keyword in the
	 * syntax, this returns the full keyword text/name.
	 * @return null if invalid token type or is not a keyword according to Proparse.
	 * @see #hasDefaultText(int)
	 */
	public static String getDefaultText(int tokenType) {
		TokenInfo info = getInfo(tokenType);
		if (info==null) return null;
		return info.keywordText;
	}


	/** Implements TokenTypesI, calls getTokenName(). */
	public String getName(int tokenType) {return getTokenName(tokenType);}
	

	/** Get the token name for this token type.
	 * @return null if invalid token type.
	 */
	public static String getTokenName(int tokenType) {
		TokenInfo info = getInfo(tokenType);
		if (info==null) return null;
		return NodeTypes.getTypeName(tokenType);
	}
	
	/** A "natural" node is one that represents a real token in the
	 * original source code. A "synthetic" token is one that was created
	 * by the parser just for syntax tree organization.
	 * By convention, a natural node has a node name that is ALLCAPS, and
	 * a synthetic node has a node name which is Mixed_Case.
	 */
	public static boolean isNatural(int tokenType) {
		TokenInfo info = getInfo(tokenType);
		if (info==null) return false;
		return info.isNatural;
	}
	
	/** Proparse tracks keywords that are part of the syntax, but does not
	 * track keywords that are only attribute/method names. DISPLAY is
	 * a keyword type, ID is not.
	 * See Proparse's node attribute documentation for NODE_TYPE_KEYWORD.
	 */
	public static boolean hasDefaultText(int tokenType) {
		TokenInfo info = getInfo(tokenType);
		if (info==null) return false;
		return (info.keywordText != null);
	}
	
	/** Is the integer a valid token type? */
	public static boolean isValid(int tokenType) {
		return (getInfo(tokenType) != null);
	}

	
	private static void assignDefaultText() {

		// Operators and punctuation
		allTokens[SLASH].keywordText = "/";
		allTokens[LEXCOLON].keywordText = ":";
		allTokens[OBJCOLON].keywordText = ":";
		allTokens[DOUBLECOLON].keywordText = "::";
		allTokens[LEXAT].keywordText = "@";
		allTokens[LEFTBRACE].keywordText = "[";
		allTokens[RIGHTBRACE].keywordText = "]";
		allTokens[CARET].keywordText = "^";
		allTokens[COMMA].keywordText = ",";
		allTokens[EXCLAMATION].keywordText = "!";
		allTokens[EQUAL].keywordText = "=";
		allTokens[LEFTPAREN].keywordText = "(";
		allTokens[RIGHTPAREN].keywordText = ")";
		allTokens[SEMI].keywordText = ";";
		allTokens[STAR].keywordText = "*";
		allTokens[UNKNOWNVALUE].keywordText = "?";
		allTokens[BACKTICK].keywordText = "`";
		allTokens[NAMEDOT].keywordText = ".";
		allTokens[PERIOD].keywordText = ".";
		allTokens[GTOREQUAL].keywordText = ">=";
		allTokens[RIGHTANGLE].keywordText = ">";
		allTokens[GTORLT].keywordText = "<>";
		allTokens[LTOREQUAL].keywordText = "<=";
		allTokens[LEFTANGLE].keywordText = "<";
		allTokens[PLUS].keywordText = "+";
		allTokens[MINUS].keywordText = "-";

		// Keywords
		allTokens[AACBIT].keywordText = "_CBIT";
		allTokens[AACONTROL].keywordText = "_CONTROL";
		allTokens[AALIST].keywordText = "_LIST";
		allTokens[AAMEMORY].keywordText = "_MEMORY";
		allTokens[AAMSG].keywordText = "_MSG";
		allTokens[AAPCONTROL].keywordText = "_PCONTROL";
		allTokens[AASERIAL].keywordText = "_SERIAL-NUM";
		allTokens[AATRACE].keywordText = "_TRACE";
		allTokens[ABSOLUTE].keywordText = "ABSOLUTE";
		allTokens[ACCELERATOR].keywordText = "ACCELERATOR";
		allTokens[ACCUMULATE].keywordText = "ACCUMULATE";
		allTokens[ACTIVEWINDOW].keywordText = "ACTIVE-WINDOW";
		allTokens[ADD].keywordText = "ADD";
		allTokens[ADVISE].keywordText = "ADVISE";
		allTokens[ALERTBOX].keywordText = "ALERT-BOX";
		allTokens[ALIAS].keywordText = "ALIAS";
		allTokens[ALL].keywordText = "ALL";
		allTokens[ALLOWREPLICATION].keywordText = "ALLOW-REPLICATION";
		allTokens[ALTER].keywordText = "ALTER";
		allTokens[ALTERNATEKEY].keywordText = "ALTERNATE-KEY";
		allTokens[AMBIGUOUS].keywordText = "AMBIGUOUS";
		allTokens[ANALYZE].keywordText = "ANALYZE";
		allTokens[AND].keywordText = "AND";
		allTokens[ANSIONLY].keywordText = "ANSI-ONLY";
		allTokens[ANY].keywordText = "ANY";
		allTokens[ANYWHERE].keywordText = "ANYWHERE";
		allTokens[APPEND].keywordText = "APPEND";
		allTokens[APPLICATION].keywordText = "APPLICATION";
		allTokens[APPLY].keywordText = "APPLY";
		allTokens[ARRAYMESSAGE].keywordText = "ARRAY-MESSAGE";
		allTokens[AS].keywordText = "AS";
		allTokens[ASC].keywordText = "ASC";
		allTokens[ASCENDING].keywordText = "ASCENDING";
		allTokens[ASKOVERWRITE].keywordText = "ASK-OVERWRITE";
		allTokens[ASSIGN].keywordText = "ASSIGN";
		allTokens[ASYNCHRONOUS].keywordText = "ASYNCHRONOUS";
		allTokens[AT].keywordText = "AT";
		allTokens[ATTACHMENT].keywordText = "ATTACHMENT";
		allTokens[ATTRSPACE].keywordText = "ATTR-SPACE";
		allTokens[AUTHORIZATION].keywordText = "AUTHORIZATION";
		allTokens[AUTOCOMPLETION].keywordText = "AUTO-COMPLETION";
		allTokens[AUTOENDKEY].keywordText = "AUTO-END-KEY";
		allTokens[AUTOGO].keywordText = "AUTO-GO";
		allTokens[AUTOMATIC].keywordText = "AUTOMATIC";
		allTokens[AUTORETURN].keywordText = "AUTO-RETURN";
		allTokens[AVAILABLE].keywordText = "AVAILABLE";
		allTokens[AVERAGE].keywordText = "AVERAGE";
		allTokens[AVG].keywordText = "AVG";
		allTokens[BACKGROUND].keywordText = "BACKGROUND";
		allTokens[BACKWARDS].keywordText = "BACKWARDS";
		allTokens[BASEKEY].keywordText = "BASE-KEY";
		allTokens[BEFOREHIDE].keywordText = "BEFORE-HIDE";
		allTokens[BEGINS].keywordText = "BEGINS";
		allTokens[BELL].keywordText = "BELL";
		allTokens[BETWEEN].keywordText = "BETWEEN";
		allTokens[BGCOLOR].keywordText = "BGCOLOR";
		allTokens[BIGENDIAN].keywordText = "BIG-ENDIAN";
		allTokens[BINARY].keywordText = "BINARY";
		allTokens[BINDWHERE].keywordText = "BIND-WHERE";
		allTokens[BLANK].keywordText = "BLANK";
		allTokens[BOTH].keywordText = "BOTH";
		allTokens[BOTTOM].keywordText = "BOTTOM";
		allTokens[BREAK].keywordText = "BREAK";
		allTokens[BROWSE].keywordText = "BROWSE";
		allTokens[BTOS].keywordText = "BTOS";
		allTokens[BUFFER].keywordText = "BUFFER";
		allTokens[BUFFERCHARS].keywordText = "BUFFER-CHARS";
		allTokens[BUFFERCOMPARE].keywordText = "BUFFER-COMPARE";
		allTokens[BUFFERCOPY].keywordText = "BUFFER-COPY";
		allTokens[BUFFERLINES].keywordText = "BUFFER-LINES";
		allTokens[BUFFERNAME].keywordText = "BUFFER-NAME";
		allTokens[BUTTON].keywordText = "BUTTON";
		allTokens[BUTTONS].keywordText = "BUTTONS";
		allTokens[BY].keywordText = "BY";
		allTokens[BYPOINTER].keywordText = "BY-POINTER";
		allTokens[BYTE].keywordText = "BYTE";
		allTokens[BYVARIANTPOINTER].keywordText = "BY-VARIANT-POINTER";
		allTokens[CACHE].keywordText = "CACHE";
		allTokens[CACHESIZE].keywordText = "CACHE-SIZE";
		allTokens[CALL].keywordText = "CALL";
		allTokens[CANDO].keywordText = "CAN-DO";
		allTokens[CANFIND].keywordText = "CAN-FIND";
		allTokens[CANQUERY].keywordText = "CAN-QUERY";
		allTokens[CANSET].keywordText = "CAN-SET";
		allTokens[CANCELBUTTON].keywordText = "CANCEL-BUTTON";
		allTokens[CAPS].keywordText = "UPPER";
		allTokens[CASE].keywordText = "CASE";
		allTokens[CASESENSITIVE].keywordText = "CASE-SENSITIVE";
		allTokens[CDECL_KW].keywordText = "CDECL";
		allTokens[CENTERED].keywordText = "CENTERED";
		allTokens[CHAINED].keywordText = "CHAINED";
		allTokens[CHARACTER].keywordText = "CHARACTER";
		allTokens[CHARACTERLENGTH].keywordText = "CHARACTERLENGTH";
		allTokens[CHECK].keywordText = "CHECK";
		allTokens[CHOOSE].keywordText = "CHOOSE";
		allTokens[CHR].keywordText = "CHR";
		allTokens[CLEAR].keywordText = "CLEAR";
		allTokens[CLIPBOARD].keywordText = "CLIPBOARD";
		allTokens[CLOSE].keywordText = "CLOSE";
		allTokens[CODEBASELOCATOR].keywordText = "CODEBASE-LOCATOR";
		allTokens[CODEPAGECONVERT].keywordText = "CODEPAGE-CONVERT";
		allTokens[COLLATE].keywordText = "COLLATE";
		allTokens[COLOF].keywordText = "COL-OF";
		allTokens[COLON].keywordText = "COLON";
		allTokens[COLONALIGNED].keywordText = "COLON-ALIGNED";
		allTokens[COLOR].keywordText = "COLOR";
		allTokens[COLORTABLE].keywordText = "COLOR-TABLE";
		allTokens[COLUMN].keywordText = "COLUMN";
		allTokens[COLUMNBGCOLOR].keywordText = "COLUMN-BGCOLOR";
		allTokens[COLUMNDCOLOR].keywordText = "COLUMN-DCOLOR";
		allTokens[COLUMNFGCOLOR].keywordText = "COLUMN-FGCOLOR";
		allTokens[COLUMNFONT].keywordText = "COLUMN-FONT";
		allTokens[COLUMNLABEL].keywordText = "COLUMN-LABEL";
		allTokens[COLUMNOF].keywordText = "COLUMN-OF";
		allTokens[COLUMNPFCOLOR].keywordText = "COLUMN-PFCOLOR";
		allTokens[COLUMNS].keywordText = "COLUMNS";
		allTokens[COMHANDLE].keywordText = "COMPONENT-HANDLE";
		allTokens[COMBOBOX].keywordText = "COMBO-BOX";
		allTokens[COMMAND].keywordText = "COMMAND";
		allTokens[COMPARE].keywordText = "COMPARE";
		allTokens[COMPARES].keywordText = "COMPARES";
		allTokens[COMPLETE].keywordText = "COMPLETE";
		allTokens[COMPILE].keywordText = "COMPILE";
		allTokens[COMPILER].keywordText = "COMPILER";
		allTokens[COMSELF].keywordText = "COM-SELF";
		allTokens[CONFIGNAME].keywordText = "CONFIG-NAME";
		allTokens[CONNECT].keywordText = "CONNECT";
		allTokens[CONNECTED].keywordText = "CONNECTED";
		allTokens[CONTAINS].keywordText = "CONTAINS";
		allTokens[CONTENTS].keywordText = "CONTENTS";
		allTokens[CONTEXT].keywordText = "CONTEXT";
		allTokens[CONTEXTHELP].keywordText = "CONTEXT-HELP";
		allTokens[CONTEXTHELPFILE].keywordText = "CONTEXT-HELP-FILE";
		allTokens[CONTEXTHELPID].keywordText = "CONTEXT-HELP-ID";
		allTokens[CONTEXTPOPUP].keywordText = "CONTEXT-POPUP";
		allTokens[CONTROL].keywordText = "CONTROL";
		allTokens[CONTROLFRAME].keywordText = "CONTROL-FRAME";
		allTokens[CONVERT].keywordText = "CONVERT";
		allTokens[CONVERT3DCOLORS].keywordText = "CONVERT-3D-COLORS";
		allTokens[COUNT].keywordText = "COUNT";
		allTokens[COUNTOF].keywordText = "COUNT-OF";
		allTokens[CREATE].keywordText = "CREATE";
		allTokens[CREATETESTFILE].keywordText = "CREATE-TEST-FILE";
		allTokens[CURRENT].keywordText = "CURRENT";
		allTokens[CURRENTCHANGED].keywordText = "CURRENT-CHANGED";
		allTokens[CURRENTENVIRONMENT].keywordText = "CURRENT-ENVIRONMENT";
		allTokens[CURRENTLANGUAGE].keywordText = "CURRENT-LANGUAGE";
		allTokens[CURRENTRESULTROW].keywordText = "CURRENT-RESULT-ROW";
		allTokens[CURRENTVALUE].keywordText = "CURRENT-VALUE";
		allTokens[CURRENTWINDOW].keywordText = "CURRENT-WINDOW";
		allTokens[CURSOR].keywordText = "CURSOR";
		allTokens[DATABASE].keywordText = "DATABASE";
		allTokens[DATABIND].keywordText = "DATA-BIND";
		allTokens[DATASERVERS].keywordText = "DATASERVERS";
		allTokens[DATE].keywordText = "DATE";
		allTokens[DAY].keywordText = "DAY";
		allTokens[DBCODEPAGE].keywordText = "DBCODEPAGE";
		allTokens[DBCOLLATION].keywordText = "DBCOLLATION";
		allTokens[DBIMS].keywordText = "DBIMS";
		allTokens[DBNAME].keywordText = "DBNAME";
		allTokens[DBPARAM].keywordText = "DBPARAM";
		allTokens[DBRESTRICTIONS].keywordText = "DBRESTRICTIONS";
		allTokens[DBTASKID].keywordText = "DBTASKID";
		allTokens[DBTYPE].keywordText = "DBTYPE";
		allTokens[DBVERSION].keywordText = "DBVERSION";
		allTokens[DCOLOR].keywordText = "DCOLOR";
		allTokens[DDE].keywordText = "DDE";
		allTokens[DEBLANK].keywordText = "DEBLANK";
		allTokens[DEBUG].keywordText = "DEBUG";
		allTokens[DEBUGLIST].keywordText = "DEBUG-LIST";
		allTokens[DEBUGGER].keywordText = "DEBUGGER";
		allTokens[DECIMAL].keywordText = "DECIMAL";
		allTokens[DECIMALS].keywordText = "DECIMALS";
		allTokens[DECLARE].keywordText = "DECLARE";
		allTokens[DEFAULT].keywordText = "DEFAULT";
		allTokens[DEFAULTBUTTON].keywordText = "DEFAULT-BUTTON";
		allTokens[DEFAULTEXTENSION].keywordText = "DEFAULT-EXTENSION";
		allTokens[DEFAULTNOXLATE].keywordText = "DEFAULT-NOXLATE";
		allTokens[DEFAULTWINDOW].keywordText = "DEFAULT-WINDOW";
		allTokens[DEFERLOBFETCH].keywordText = "DEFER-LOB-FETCH";
		allTokens[DEFINE].keywordText = "DEFINE";
		allTokens[DEFINED].keywordText = "DEFINED";
		allTokens[DELETE_KW].keywordText = "DELETE";
		allTokens[DELETERESULTLISTENTRY].keywordText = "DELETE-RESULT-LIST-ENTRY";
		allTokens[DELIMITER].keywordText = "DELIMITER";
		allTokens[DESCENDING].keywordText = "DESCENDING";
		allTokens[DESELECTION].keywordText = "DESELECTION";
		allTokens[DIALOGBOX].keywordText = "DIALOG-BOX";
		allTokens[DIALOGHELP].keywordText = "DIALOG-HELP";
		allTokens[DICTIONARY].keywordText = "DICTIONARY";
		allTokens[DIR].keywordText = "DIR";
		allTokens[DISABLE].keywordText = "DISABLE";
		allTokens[DISABLEAUTOZAP].keywordText = "DISABLE-AUTO-ZAP";
		allTokens[DISABLED].keywordText = "DISABLED";
		allTokens[DISCONNECT].keywordText = "DISCONNECT";
		allTokens[DISPLAY].keywordText = "DISPLAY";
		allTokens[DISTINCT].keywordText = "DISTINCT";
		allTokens[DO].keywordText = "DO";
		allTokens[DOS].keywordText = "DOS";
		allTokens[DOUBLE].keywordText = "DOUBLE";
		allTokens[DOWN].keywordText = "DOWN";
		allTokens[DROP].keywordText = "DROP";
		allTokens[DROPDOWN].keywordText = "DROP-DOWN";
		allTokens[DROPDOWNLIST].keywordText = "DROP-DOWN-LIST";
		allTokens[DROPFILENOTIFY].keywordText = "DROP-FILE-NOTIFY";
		allTokens[DROPTARGET].keywordText = "DROP-TARGET";
		allTokens[DUMP].keywordText = "DUMP";
		allTokens[DYNAMIC].keywordText = "DYNAMIC";
		allTokens[DYNAMICFUNCTION].keywordText = "DYNAMIC-FUNCTION";
		allTokens[EACH].keywordText = "EACH";
		allTokens[ECHO].keywordText = "ECHO";
		allTokens[EDGECHARS].keywordText = "EDGE-CHARS";
		allTokens[EDGEPIXELS].keywordText = "EDGE-PIXELS";
		allTokens[EDITUNDO].keywordText = "EDIT-UNDO";
		allTokens[EDITING].keywordText = "EDITING";
		allTokens[EDITOR].keywordText = "EDITOR";
		allTokens[ELSE].keywordText = "ELSE";
		allTokens[EMPTY].keywordText = "EMPTY";
		allTokens[ENABLE].keywordText = "ENABLE";
		allTokens[ENCODE].keywordText = "ENCODE";
		allTokens[END].keywordText = "END";
		allTokens[ENDMOVE].keywordText = "END-MOVE";
		allTokens[ENDRESIZE].keywordText = "END-RESIZE";
		allTokens[ENDROWRESIZE].keywordText = "END-ROW-RESIZE";
		allTokens[ENDKEY].keywordText = "END-KEY";
		allTokens[ENTERED].keywordText = "ENTERED";
		allTokens[ENTRY].keywordText = "ENTRY";
		allTokens[EQ].keywordText = "EQ";
		allTokens[ERROR].keywordText = "ERROR";
		allTokens[ERRORSTATUS].keywordText = "ERROR-STATUS";
		allTokens[ESCAPE].keywordText = "ESCAPE";
		allTokens[ETIME_KW].keywordText = "ETIME";
		allTokens[EVENTPROCEDURE].keywordText = "EVENT-PROCEDURE";
		allTokens[EVENTS].keywordText = "EVENTS";
		allTokens[EXCEPT].keywordText = "EXCEPT";
		allTokens[EXCLUSIVEID].keywordText = "EXCLUSIVE-ID";
		allTokens[EXCLUSIVELOCK].keywordText = "EXCLUSIVE-LOCK";
		allTokens[EXCLUSIVEWEBUSER].keywordText = "EXCLUSIVE-WEB-USER";
		allTokens[EXECUTE].keywordText = "EXECUTE";
		allTokens[EXISTS].keywordText = "EXISTS";
		allTokens[EXP].keywordText = "EXP";
		allTokens[EXPAND].keywordText = "EXPAND";
		allTokens[EXPANDABLE].keywordText = "EXPANDABLE";
		allTokens[EXPLICIT].keywordText = "EXPLICIT";
		allTokens[EXPORT].keywordText = "EXPORT";
		allTokens[EXTENDED].keywordText = "EXTENDED";
		allTokens[EXTENT].keywordText = "EXTENT";
		allTokens[EXTERNAL].keywordText = "EXTERNAL";
		allTokens[FALSE_KW].keywordText = "FALSE";
		allTokens[FETCH].keywordText = "FETCH";
		allTokens[FGCOLOR].keywordText = "FGCOLOR";
		allTokens[FIELD].keywordText = "FIELD";
		allTokens[FIELDS].keywordText = "FIELDS";
		allTokens[FILE].keywordText = "FILE-NAME";
		allTokens[FILEINFORMATION].keywordText = "FILE-INFORMATION";
		allTokens[FILL].keywordText = "FILL";
		allTokens[FILLIN].keywordText = "FILL-IN";
		allTokens[FILTERS].keywordText = "FILTERS";
		allTokens[FIND].keywordText = "FIND";
		allTokens[FINDCASESENSITIVE].keywordText = "FIND-CASE-SENSITIVE";
		allTokens[FINDER].keywordText = "FINDER";
		allTokens[FINDGLOBAL].keywordText = "FIND-GLOBAL";
		allTokens[FINDNEXTOCCURRENCE].keywordText = "FIND-NEXT-OCCURRENCE";
		allTokens[FINDPREVOCCURRENCE].keywordText = "FIND-PREV-OCCURRENCE";
		allTokens[FINDSELECT].keywordText = "FIND-SELECT";
		allTokens[FINDWRAPAROUND].keywordText = "FIND-WRAP-AROUND";
		allTokens[FIRST].keywordText = "FIRST";
		allTokens[FIRSTOF].keywordText = "FIRST-OF";
		allTokens[FITLASTCOLUMN].keywordText = "FIT-LAST-COLUMN";
		allTokens[FIXEDONLY].keywordText = "FIXED-ONLY";
		allTokens[FLATBUTTON].keywordText = "FLAT-BUTTON";
		allTokens[FLOAT].keywordText = "FLOAT";
		allTokens[FOCUS].keywordText = "FOCUS";
		allTokens[FONT].keywordText = "FONT";
		allTokens[FONTBASEDLAYOUT].keywordText = "FONT-BASED-LAYOUT";
		allTokens[FONTTABLE].keywordText = "FONT-TABLE";
		allTokens[FOR].keywordText = "FOR";
		allTokens[FORCEFILE].keywordText = "FORCE-FILE";
		allTokens[FORMINPUT].keywordText = "FORMINPUT";
		allTokens[FORMAT].keywordText = "FORMAT";
		allTokens[FORWARDS].keywordText = "FORWARDS";
		allTokens[FRAME].keywordText = "FRAME";
		allTokens[FRAMECOL].keywordText = "FRAME-COL";
		allTokens[FRAMEDB].keywordText = "FRAME-DB";
		allTokens[FRAMEDOWN].keywordText = "FRAME-DOWN";
		allTokens[FRAMEFIELD].keywordText = "FRAME-FIELD";
		allTokens[FRAMEFILE].keywordText = "FRAME-FILE";
		allTokens[FRAMEINDEX].keywordText = "FRAME-INDEX";
		allTokens[FRAMELINE].keywordText = "FRAME-LINE";
		allTokens[FRAMENAME].keywordText = "FRAME-NAME";
		allTokens[FRAMEROW].keywordText = "FRAME-ROW";
		allTokens[FRAMEVALUE].keywordText = "FRAME-VALUE";
		allTokens[FREQUENCY].keywordText = "FREQUENCY";
		allTokens[FROM].keywordText = "FROM";
		allTokens[FROMCURRENT].keywordText = "FROM-CURRENT";
		allTokens[FUNCTION].keywordText = "FUNCTION";
		allTokens[GE].keywordText = "GE";
		allTokens[GENERATEMD5].keywordText = "GENERATE-MD5";
		allTokens[GET].keywordText = "GET";
		allTokens[GETBITS].keywordText = "GET-BITS";
		allTokens[GETBUFFERHANDLE].keywordText = "GET-BUFFER-HANDLE";
		allTokens[GETBYTE].keywordText = "GET-BYTE";
		allTokens[GETBYTES].keywordText = "GET-BYTES";
		allTokens[GETBYTEORDER].keywordText = "GET-BYTE-ORDER";
		allTokens[GETCGILIST].keywordText = "GET-CGI-LIST";
		allTokens[GETCGIVALUE].keywordText = "GET-CGI-VALUE";
		allTokens[GETCODEPAGES].keywordText = "GET-CODEPAGES";
		allTokens[GETCOLLATIONS].keywordText = "GET-COLLATIONS";
		allTokens[GETCONFIGVALUE].keywordText = "GET-CONFIG-VALUE";
		allTokens[GETDOUBLE].keywordText = "GET-DOUBLE";
		allTokens[GETFILE].keywordText = "GET-FILE";
		allTokens[GETFLOAT].keywordText = "GET-FLOAT";
		allTokens[GETKEYVALUE].keywordText = "GET-KEY-VALUE";
		allTokens[GETLICENSE].keywordText = "GET-LICENSE";
		allTokens[GETLONG].keywordText = "GET-LONG";
		allTokens[GETPOINTERVALUE].keywordText = "GET-POINTER-VALUE";
		allTokens[GETSHORT].keywordText = "GET-SHORT";
		allTokens[GETSIZE].keywordText = "GET-SIZE";
		allTokens[GETSTRING].keywordText = "GET-STRING";
		allTokens[GETUNSIGNEDSHORT].keywordText = "GET-UNSIGNED-SHORT";
		allTokens[GLOBAL].keywordText = "GLOBAL";
		allTokens[GOON].keywordText = "GO-ON";
		allTokens[GOPENDING].keywordText = "GO-PENDING";
		allTokens[GRANT].keywordText = "GRANT";
		allTokens[GRAPHICEDGE].keywordText = "GRAPHIC-EDGE";
		allTokens[GROUP].keywordText = "GROUP";
		allTokens[GTHAN].keywordText = "GT";
		allTokens[HANDLE].keywordText = "HANDLE";
		allTokens[HAVING].keywordText = "HAVING";
		allTokens[HEADER].keywordText = "HEADER";
		allTokens[HEIGHT].keywordText = "HEIGHT";
		allTokens[HELP].keywordText = "HELP";
		allTokens[HELPTOPIC].keywordText = "HELP-TOPIC";
		allTokens[HIDE].keywordText = "HIDE";
		allTokens[HINT].keywordText = "HINT";
		allTokens[HORIZONTAL].keywordText = "HORIZONTAL";
		allTokens[HOSTBYTEORDER].keywordText = "HOST-BYTE-ORDER";
		allTokens[HTMLENDOFLINE].keywordText = "HTML-END-OF-LINE";
		allTokens[HTMLFRAMEBEGIN].keywordText = "HTML-FRAME-BEGIN";
		allTokens[HTMLFRAMEEND].keywordText = "HTML-FRAME-END";
		allTokens[HTMLHEADERBEGIN].keywordText = "HTML-HEADER-BEGIN";
		allTokens[HTMLHEADEREND].keywordText = "HTML-HEADER-END";
		allTokens[HTMLTITLEBEGIN].keywordText = "HTML-TITLE-BEGIN";
		allTokens[HTMLTITLEEND].keywordText = "HTML-TITLE-END";
		allTokens[IF].keywordText = "IF";
		allTokens[IMAGE].keywordText = "IMAGE";
		allTokens[IMAGEDOWN].keywordText = "IMAGE-DOWN";
		allTokens[IMAGEINSENSITIVE].keywordText = "IMAGE-INSENSITIVE";
		allTokens[IMAGESIZE].keywordText = "IMAGE-SIZE";
		allTokens[IMAGESIZECHARS].keywordText = "IMAGE-SIZE-CHARS";
		allTokens[IMAGESIZEPIXELS].keywordText = "IMAGE-SIZE-PIXELS";
		allTokens[IMAGEUP].keywordText = "IMAGE-UP";
		allTokens[IMPORT].keywordText = "IMPORT";
		allTokens[IN_KW].keywordText = "IN";
		allTokens[INCREMENTEXCLUSIVEID].keywordText = "INCREMENT-EXCLUSIVE-ID";
		allTokens[INDEX].keywordText = "INDEX";
		allTokens[INDEXHINT].keywordText = "INDEX-HINT";
		allTokens[INDEXEDREPOSITION].keywordText = "INDEXED-REPOSITION";
		allTokens[INDICATOR].keywordText = "INDICATOR";
		allTokens[INFORMATION].keywordText = "INFORMATION";
		allTokens[INITIAL].keywordText = "INITIAL";
		allTokens[INITIALDIR].keywordText = "INITIAL-DIR";
		allTokens[INITIALFILTER].keywordText = "INITIAL-FILTER";
		allTokens[INITIATE].keywordText = "INITIATE";
		allTokens[INNER].keywordText = "INNER";
		allTokens[INNERCHARS].keywordText = "INNER-CHARS";
		allTokens[INNERLINES].keywordText = "INNER-LINES";
		allTokens[INPUT].keywordText = "INPUT";
		allTokens[INPUTOUTPUT].keywordText = "INPUT-OUTPUT";
		allTokens[INSERT].keywordText = "INSERT";
		allTokens[INTEGER].keywordText = "INTEGER";
		allTokens[INTO].keywordText = "INTO";
		allTokens[IS].keywordText = "IS";
		allTokens[ISATTRSPACE].keywordText = "IS-ATTR-SPACE";
		allTokens[ISLEADBYTE].keywordText = "IS-LEAD-BYTE";
		allTokens[ITEM].keywordText = "ITEM";
		allTokens[JOIN].keywordText = "JOIN";
		allTokens[JOINBYSQLDB].keywordText = "JOIN-BY-SQLDB";
		allTokens[KBLABEL].keywordText = "KBLABEL";
		allTokens[KEEPMESSAGES].keywordText = "KEEP-MESSAGES";
		allTokens[KEEPTABORDER].keywordText = "KEEP-TAB-ORDER";
		allTokens[KEY].keywordText = "KEY";
		allTokens[KEYCODE].keywordText = "KEY-CODE";
		allTokens[KEYFUNCTION].keywordText = "KEY-FUNCTION";
		allTokens[KEYLABEL].keywordText = "KEY-LABEL";
		allTokens[KEYS].keywordText = "KEYS";
		allTokens[KEYWORD].keywordText = "KEYWORD";
		allTokens[KEYWORDALL].keywordText = "KEYWORD-ALL";
		allTokens[LABEL].keywordText = "LABEL";
		allTokens[LABELBGCOLOR].keywordText = "LABEL-BGCOLOR";
		allTokens[LABELDCOLOR].keywordText = "LABEL-DCOLOR";
		allTokens[LABELFGCOLOR].keywordText = "LABEL-FGCOLOR";
		allTokens[LABELFONT].keywordText = "LABEL-FONT";
		allTokens[LANDSCAPE].keywordText = "LANDSCAPE";
		allTokens[LANGUAGES].keywordText = "LANGUAGES";
		allTokens[LARGE].keywordText = "LARGE";
		allTokens[LARGETOSMALL].keywordText = "LARGE-TO-SMALL";
		allTokens[LAST].keywordText = "LAST";
		allTokens[LASTEVENT].keywordText = "LAST-EVENT";
		allTokens[LASTOF].keywordText = "LAST-OF";
		allTokens[LASTKEY].keywordText = "LAST-KEY";
		allTokens[LC].keywordText = "LOWER";
		allTokens[LDBNAME].keywordText = "LDBNAME";
		allTokens[LE].keywordText = "LE";
		allTokens[LEAVE].keywordText = "LEAVE";
		allTokens[LEFT].keywordText = "LEFT";
		allTokens[LEFTALIGNED].keywordText = "LEFT-ALIGNED";
		allTokens[LEFTTRIM].keywordText = "LEFT-TRIM";
		allTokens[LENGTH].keywordText = "LENGTH";
		allTokens[LIBRARY].keywordText = "LIBRARY";
		allTokens[LIKE].keywordText = "LIKE";
		allTokens[LINECOUNTER].keywordText = "LINE-COUNTER";
		allTokens[LISTEVENTS].keywordText = "LIST-EVENTS";
		allTokens[LISTITEMPAIRS].keywordText = "LIST-ITEM-PAIRS";
		allTokens[LISTITEMS].keywordText = "LIST-ITEMS";
		allTokens[LISTQUERYATTRS].keywordText = "LIST-QUERY-ATTRS";
		allTokens[LISTSETATTRS].keywordText = "LIST-SET-ATTRS";
		allTokens[LISTWIDGETS].keywordText = "LIST-WIDGETS";
		allTokens[LISTING].keywordText = "LISTING";
		allTokens[LITTLEENDIAN].keywordText = "LITTLE-ENDIAN";
		allTokens[LOAD].keywordText = "LOAD";
		allTokens[LOADPICTURE].keywordText = "LOAD-PICTURE";
		allTokens[LOCKED].keywordText = "LOCKED";
		allTokens[LOG].keywordText = "LOG";
		allTokens[LOGICAL].keywordText = "LOGICAL";
		allTokens[LONG].keywordText = "LONG";
		allTokens[LOOKAHEAD].keywordText = "LOOKAHEAD";
		allTokens[LOOKUP].keywordText = "LOOKUP";
		allTokens[LTHAN].keywordText = "LT";
		allTokens[MACHINECLASS].keywordText = "MACHINE-CLASS";
		allTokens[MAP].keywordText = "MAP";
		allTokens[MARGINEXTRA].keywordText = "MARGIN-EXTRA";
		allTokens[MATCHES].keywordText = "MATCHES";
		allTokens[MAXCHARS].keywordText = "MAX-CHARS";
		allTokens[MAXROWS].keywordText = "MAX-ROWS";
		allTokens[MAXSIZE].keywordText = "MAX-SIZE";
		allTokens[MAXVALUE].keywordText = "MAX-VALUE";
		allTokens[MAXIMIZE].keywordText = "MAXIMIZE";
		allTokens[MAXIMUM].keywordText = "MAXIMUM";
		allTokens[MEMBER].keywordText = "MEMBER";
		allTokens[MEMPTR].keywordText = "MEMPTR";
		allTokens[MENU].keywordText = "MENU";
		allTokens[MENUITEM].keywordText = "MENU-ITEM";
		allTokens[MENUBAR].keywordText = "MENU-BAR";
		allTokens[MESSAGE].keywordText = "MESSAGE";
		allTokens[MESSAGELINE].keywordText = "MESSAGE-LINE";
		allTokens[MESSAGELINES].keywordText = "MESSAGE-LINES";
		allTokens[MINSIZE].keywordText = "MIN-SIZE";
		allTokens[MINVALUE].keywordText = "MIN-VALUE";
		allTokens[MINIMUM].keywordText = "MINIMUM";
		allTokens[MODULO].keywordText = "MODULO";
		allTokens[MONTH].keywordText = "MONTH";
		allTokens[MOUSE].keywordText = "MOUSE";
		allTokens[MOUSEPOINTER].keywordText = "MOUSE-POINTER";
		allTokens[MPE].keywordText = "MPE";
		allTokens[MULTIPLE].keywordText = "MULTIPLE";
		allTokens[MULTIPLEKEY].keywordText = "MULTIPLE-KEY";
		allTokens[MUSTEXIST].keywordText = "MUST-EXIST";
		allTokens[NATIVE].keywordText = "NATIVE";
		allTokens[NE].keywordText = "NE";
		allTokens[NEW].keywordText = "NEW";
		allTokens[NEXT].keywordText = "NEXT";
		allTokens[NEXTPROMPT].keywordText = "NEXT-PROMPT";
		allTokens[NEXTVALUE].keywordText = "NEXT-VALUE";
		allTokens[NO].keywordText = "NO";
		allTokens[NOAPPLY].keywordText = "NO-APPLY";
		allTokens[NOARRAYMESSAGE].keywordText = "NO-ARRAY-MESSAGE";
		allTokens[NOASSIGN].keywordText = "NO-ASSIGN";
		allTokens[NOATTRLIST].keywordText = "NO-ATTR-LIST";
		allTokens[NOATTRSPACE].keywordText = "NO-ATTR-SPACE";
		allTokens[NOAUTOVALIDATE].keywordText = "NO-AUTO-VALIDATE";
		allTokens[NOBINDWHERE].keywordText = "NO-BIND-WHERE";
		allTokens[NOBOX].keywordText = "NO-BOX";
		allTokens[NOCOLUMNSCROLLING].keywordText = "NO-COLUMN-SCROLLING";
		allTokens[NOCONSOLE].keywordText = "NO-CONSOLE";
		allTokens[NOCONVERT].keywordText = "NO-CONVERT";
		allTokens[NOCONVERT3DCOLORS].keywordText = "NO-CONVERT-3D-COLORS";
		allTokens[NOCURRENTVALUE].keywordText = "NO-CURRENT-VALUE";
		allTokens[NODEBUG].keywordText = "NO-DEBUG";
		allTokens[NODRAG].keywordText = "NO-DRAG";
		allTokens[NOECHO].keywordText = "NO-ECHO";
		allTokens[NOEMPTYSPACE].keywordText = "NO-EMPTY-SPACE";
		allTokens[NOERROR_KW].keywordText = "NO-ERROR";
		allTokens[NOFILL].keywordText = "NO-FILL";
		allTokens[NOFOCUS].keywordText = "NO-FOCUS";
		allTokens[NOHELP].keywordText = "NO-HELP";
		allTokens[NOHIDE].keywordText = "NO-HIDE";
		allTokens[NOINDEXHINT].keywordText = "NO-INDEX-HINT";
		allTokens[NOJOINBYSQLDB].keywordText = "NO-JOIN-BY-SQLDB";
		allTokens[NOLABELS].keywordText = "NO-LABELS";
		allTokens[NOLOCK].keywordText = "NO-LOCK";
		allTokens[NOLOOKAHEAD].keywordText = "NO-LOOKAHEAD";
		allTokens[NOMAP].keywordText = "NO-MAP";
		allTokens[NOMESSAGE].keywordText = "NO-MESSAGE";
		allTokens[NONE].keywordText = "NONE";
		allTokens[NOPAUSE].keywordText = "NO-PAUSE";
		allTokens[NOPREFETCH].keywordText = "NO-PREFETCH";
		allTokens[NORETURNVALUE].keywordText = "NO-RETURN-VALUE";
		allTokens[NORMAL].keywordText = "NORMAL";
		allTokens[NOROWMARKERS].keywordText = "NO-ROW-MARKERS";
		allTokens[NOSCROLLBARVERTICAL].keywordText = "NO-SCROLLBAR-VERTICAL";
		allTokens[NOSEPARATECONNECTION].keywordText = "NO-SEPARATE-CONNECTION";
		allTokens[NOSEPARATORS].keywordText = "NO-SEPARATORS";
		allTokens[NOTABSTOP].keywordText = "NO-TAB-STOP";
		allTokens[NOUNDERLINE].keywordText = "NO-UNDERLINE";
		allTokens[NOUNDO].keywordText = "NO-UNDO";
		allTokens[NOVALIDATE].keywordText = "NO-VALIDATE";
		allTokens[NOWAIT].keywordText = "NO-WAIT";
		allTokens[NOWORDWRAP].keywordText = "NO-WORD-WRAP";
		allTokens[NOT].keywordText = "NOT";
		allTokens[NULL_KW].keywordText = "NULL";
		allTokens[NUMALIASES].keywordText = "NUM-ALIASES";
		allTokens[NUMCOPIES].keywordText = "NUM-COPIES";
		allTokens[NUMDBS].keywordText = "NUM-DBS";
		allTokens[NUMENTRIES].keywordText = "NUM-ENTRIES";
		allTokens[NUMRESULTS].keywordText = "NUM-RESULTS";
		allTokens[NUMERIC].keywordText = "NUMERIC";
		allTokens[OBJECT].keywordText = "OBJECT";
		allTokens[OCTETLENGTH].keywordText = "OCTET-LENGTH";
		allTokens[OF].keywordText = "OF";
		allTokens[OFF].keywordText = "OFF";
		allTokens[OK].keywordText = "OK";
		allTokens[OKCANCEL].keywordText = "OK-CANCEL";
		allTokens[OLD].keywordText = "OLD";
		allTokens[ON].keywordText = "ON";
		allTokens[ONLY].keywordText = "ONLY";
		allTokens[OPEN].keywordText = "OPEN";
		allTokens[OPSYS].keywordText = "OPSYS";
		allTokens[OPTION].keywordText = "OPTION";
		allTokens[OR].keywordText = "OR";
		allTokens[ORDER].keywordText = "ORDER";
		allTokens[ORDEREDJOIN].keywordText = "ORDERED-JOIN";
		allTokens[ORDINAL].keywordText = "ORDINAL";
		allTokens[OS2].keywordText = "OS2";
		allTokens[OS400].keywordText = "OS400";
		allTokens[OSAPPEND].keywordText = "OS-APPEND";
		allTokens[OSCOMMAND].keywordText = "OS-COMMAND";
		allTokens[OSCOPY].keywordText = "OS-COPY";
		allTokens[OSCREATEDIR].keywordText = "OS-CREATE-DIR";
		allTokens[OSDELETE].keywordText = "OS-DELETE";
		allTokens[OSDIR].keywordText = "OS-DIR";
		allTokens[OSDRIVES].keywordText = "OS-DRIVES";
		allTokens[OSERROR].keywordText = "OS-ERROR";
		allTokens[OSGETENV].keywordText = "OS-GETENV";
		allTokens[OSRENAME].keywordText = "OS-RENAME";
		allTokens[OTHERWISE].keywordText = "OTHERWISE";
		allTokens[OUTER].keywordText = "OUTER";
		allTokens[OUTERJOIN].keywordText = "OUTER-JOIN";
		allTokens[OUTPUT].keywordText = "OUTPUT";
		allTokens[OVERLAY].keywordText = "OVERLAY";
		allTokens[OVERRIDE].keywordText = "OVERRIDE";
		allTokens[PAGE].keywordText = "PAGE";
		allTokens[PAGEBOTTOM].keywordText = "PAGE-BOTTOM";
		allTokens[PAGENUMBER].keywordText = "PAGE-NUMBER";
		allTokens[PAGESIZE_KW].keywordText = "PAGE-SIZE";
		allTokens[PAGETOP].keywordText = "PAGE-TOP";
		allTokens[PAGEWIDTH].keywordText = "PAGE-WIDTH";
		allTokens[PAGED].keywordText = "PAGED";
		allTokens[PARAMETER].keywordText = "PARAMETER";
		allTokens[PARENT].keywordText = "PARENT";
		allTokens[PARTIALKEY].keywordText = "PARTIAL-KEY";
		allTokens[PASCAL_KW].keywordText = "PASCAL";
		allTokens[PAUSE].keywordText = "PAUSE";
		allTokens[PDBNAME].keywordText = "PDBNAME";
		allTokens[PERFORMANCE].keywordText = "PERFORMANCE";
		allTokens[PERSISTENT].keywordText = "PERSISTENT";
		allTokens[PFCOLOR].keywordText = "PFCOLOR";
		allTokens[PINNABLE].keywordText = "PINNABLE";
		allTokens[PORTRAIT].keywordText = "PORTRAIT";
		allTokens[POSITION].keywordText = "POSITION";
		allTokens[PRECISION].keywordText = "PRECISION";
		allTokens[PREPROCESS].keywordText = "PREPROCESS";
		allTokens[PRESELECT].keywordText = "PRESELECT";
		allTokens[PREV].keywordText = "PREV";
		allTokens[PRIMARY].keywordText = "PRIMARY";
		allTokens[PRINTER].keywordText = "PRINTER";
		allTokens[PRINTERSETUP].keywordText = "PRINTER-SETUP";
		allTokens[PRIVATE].keywordText = "PRIVATE";
		allTokens[PRIVILEGES].keywordText = "PRIVILEGES";
		allTokens[PROCTEXT].keywordText = "PROC-TEXT";
		allTokens[PROCTEXTBUFFER].keywordText = "PROC-TEXT-BUFFER";
		allTokens[PROCHANDLE].keywordText = "PROC-HANDLE";
		allTokens[PROCSTATUS].keywordText = "PROC-STATUS";
		allTokens[PROCEDURE].keywordText = "PROCEDURE";
		allTokens[PROCESS].keywordText = "PROCESS";
		allTokens[PROFILER].keywordText = "PROFILER";
		allTokens[PROGRAMNAME].keywordText = "PROGRAM-NAME";
		allTokens[PROGRESS].keywordText = "PROGRESS";
		allTokens[PROMPT].keywordText = "PROMPT";
		allTokens[PROMPTFOR].keywordText = "PROMPT-FOR";
		allTokens[PROMSGS].keywordText = "PROMSGS";
		allTokens[PROPATH].keywordText = "PROPATH";
		allTokens[PROVERSION].keywordText = "PROVERSION";
		allTokens[PUBLIC].keywordText = "PUBLIC";
		allTokens[PUBLISH].keywordText = "PUBLISH";
		allTokens[PUT].keywordText = "PUT";
		allTokens[PUTBITS].keywordText = "PUT-BITS";
		allTokens[PUTBYTE].keywordText = "PUT-BYTE";
		allTokens[PUTBYTES].keywordText = "PUT-BYTES";
		allTokens[PUTDOUBLE].keywordText = "PUT-DOUBLE";
		allTokens[PUTFLOAT].keywordText = "PUT-FLOAT";
		allTokens[PUTKEYVALUE].keywordText = "PUT-KEY-VALUE";
		allTokens[PUTLONG].keywordText = "PUT-LONG";
		allTokens[PUTSHORT].keywordText = "PUT-SHORT";
		allTokens[PUTSTRING].keywordText = "PUT-STRING";
		allTokens[PUTUNSIGNEDSHORT].keywordText = "PUT-UNSIGNED-SHORT";
		allTokens[QUERY].keywordText = "QUERY";
		allTokens[QUERYCLOSE].keywordText = "QUERY-CLOSE";
		allTokens[QUERYOFFEND].keywordText = "QUERY-OFF-END";
		allTokens[QUERYTUNING].keywordText = "QUERY-TUNING";
		allTokens[QUESTION].keywordText = "QUESTION";
		allTokens[QUIT].keywordText = "QUIT";
		allTokens[QUOTER].keywordText = "QUOTER";
		allTokens[RINDEX].keywordText = "R-INDEX";
		allTokens[RADIOBUTTONS].keywordText = "RADIO-BUTTONS";
		allTokens[RADIOSET].keywordText = "RADIO-SET";
		allTokens[RANDOM].keywordText = "RANDOM";
		allTokens[RAW].keywordText = "RAW";
		allTokens[RAWTRANSFER].keywordText = "RAW-TRANSFER";
		allTokens[RCODEINFORMATION].keywordText = "RCODE-INFORMATION";
		allTokens[READ].keywordText = "READ";
		allTokens[READAVAILABLE].keywordText = "READ-AVAILABLE";
		allTokens[READEXACTNUM].keywordText = "READ-EXACT-NUM";
		allTokens[READONLY].keywordText = "READ-ONLY";
		allTokens[READKEY].keywordText = "READKEY";
		allTokens[REAL].keywordText = "REAL";
		allTokens[RECID].keywordText = "RECID";
		allTokens[RECORDLENGTH].keywordText = "RECORD-LENGTH";
		allTokens[RECTANGLE].keywordText = "RECTANGLE";
		allTokens[RECURSIVE].keywordText = "RECURSIVE";
		allTokens[RELEASE].keywordText = "RELEASE";
		allTokens[REPEAT].keywordText = "REPEAT";
		allTokens[REPLACE].keywordText = "REPLACE";
		allTokens[REPLICATIONCREATE].keywordText = "REPLICATION-CREATE";
		allTokens[REPLICATIONDELETE].keywordText = "REPLICATION-DELETE";
		allTokens[REPLICATIONWRITE].keywordText = "REPLICATION-WRITE";
		allTokens[REPOSITION].keywordText = "REPOSITION";
		allTokens[REPOSITIONFORWARD].keywordText = "REPOSITION-FORWARD";
		allTokens[REPOSITIONBACKWARD].keywordText = "REPOSITION-BACKWARD";
		allTokens[REPOSITIONTOROW].keywordText = "REPOSITION-TO-ROW";
		allTokens[REPOSITIONTOROWID].keywordText = "REPOSITION-TO-ROWID";
		allTokens[REQUEST].keywordText = "REQUEST";
		allTokens[RESULT].keywordText = "RESULT";
		allTokens[RETAIN].keywordText = "RETAIN";
		allTokens[RETAINSHAPE].keywordText = "RETAIN-SHAPE";
		allTokens[RETRY].keywordText = "RETRY";
		allTokens[RETRYCANCEL].keywordText = "RETRY-CANCEL";
		allTokens[RETURN].keywordText = "RETURN";
		allTokens[RETURNTOSTARTDIR].keywordText = "RETURN-TO-START-DIR";
		allTokens[RETURNVALUE].keywordText = "RETURN-VALUE";
		allTokens[RETURNS].keywordText = "RETURNS";
		allTokens[REVERSEFROM].keywordText = "REVERSE-FROM";
		allTokens[REVERT].keywordText = "REVERT";
		allTokens[REVOKE].keywordText = "REVOKE";
		allTokens[RGBVALUE].keywordText = "RGB-VALUE";
		allTokens[RIGHT].keywordText = "RIGHT";
		allTokens[RIGHTALIGNED].keywordText = "RIGHT-ALIGNED";
		allTokens[RIGHTTRIM].keywordText = "RIGHT-TRIM";
		allTokens[ROUND].keywordText = "ROUND";
		allTokens[ROW].keywordText = "ROW";
		allTokens[ROWHEIGHTCHARS].keywordText = "ROW-HEIGHT-CHARS";
		allTokens[ROWHEIGHTPIXELS].keywordText = "ROW-HEIGHT-PIXELS";
		allTokens[ROWID].keywordText = "ROWID";
		allTokens[ROWOF].keywordText = "ROW-OF";
		allTokens[RULE].keywordText = "RULE";
		allTokens[RUN].keywordText = "RUN";
		allTokens[RUNPROCEDURE].keywordText = "RUN-PROCEDURE";
		allTokens[SAVE].keywordText = "SAVE";
		allTokens[SAVECACHE].keywordText = "SAVECACHE";
		allTokens[SAVEAS].keywordText = "SAVE-AS";
		allTokens[SAXREADER].keywordText = "SAX-READER";
		allTokens[SCHEMA].keywordText = "SCHEMA";
		allTokens[SCREEN].keywordText = "SCREEN";
		allTokens[SCREENIO].keywordText = "SCREEN-IO";
		allTokens[SCREENLINES].keywordText = "SCREEN-LINES";
		allTokens[SCROLL].keywordText = "SCROLL";
		allTokens[SCROLLABLE].keywordText = "SCROLLABLE";
		allTokens[SCROLLBARHORIZONTAL].keywordText = "SCROLLBAR-HORIZONTAL";
		allTokens[SCROLLBARVERTICAL].keywordText = "SCROLLBAR-VERTICAL";
		allTokens[SCROLLING].keywordText = "SCROLLING";
		allTokens[SDBNAME].keywordText = "SDBNAME";
		allTokens[SEARCH].keywordText = "SEARCH";
		allTokens[SEARCHSELF].keywordText = "SEARCH-SELF";
		allTokens[SEARCHTARGET].keywordText = "SEARCH-TARGET";
		allTokens[SECTION].keywordText = "SECTION";
		allTokens[SEEK].keywordText = "SEEK";
		allTokens[SELECT].keywordText = "SELECT";
		allTokens[SELECTION].keywordText = "SELECTION";
		allTokens[SELECTIONLIST].keywordText = "SELECTION-LIST";
		allTokens[SELF].keywordText = "SELF";
		allTokens[SEND].keywordText = "SEND";
		allTokens[SENDSQLSTATEMENT].keywordText = "SEND-SQL-STATEMENT";
		allTokens[SEPARATECONNECTION].keywordText = "SEPARATE-CONNECTION";
		allTokens[SEPARATORS].keywordText = "SEPARATORS";
		allTokens[SERVER].keywordText = "SERVER";
		allTokens[SERVERSOCKET].keywordText = "SERVER-SOCKET";
		allTokens[SESSION].keywordText = "SESSION";
		allTokens[SET].keywordText = "SET";
		allTokens[SETBYTEORDER].keywordText = "SET-BYTE-ORDER";
		allTokens[SETCONTENTS].keywordText = "SET-CONTENTS";
		allTokens[SETCURRENTVALUE].keywordText = "SET-CURRENT-VALUE";
		allTokens[SETPOINTERVALUE].keywordText = "SET-POINTER-VALUE";
		allTokens[SETSIZE].keywordText = "SET-SIZE";
		allTokens[SETUSERID].keywordText = "SETUSERID";
		allTokens[SHARELOCK].keywordText = "SHARE-LOCK";
		allTokens[SHARED].keywordText = "SHARED";
		allTokens[SHOWSTATS].keywordText = "SHOW-STATS";
		allTokens[SIDELABELS].keywordText = "SIDE-LABELS";
		allTokens[SILENT].keywordText = "SILENT";
		allTokens[SIMPLE].keywordText = "SIMPLE";
		allTokens[SINGLE].keywordText = "SINGLE";
		allTokens[SIZE].keywordText = "SIZE";
		allTokens[SIZECHARS].keywordText = "SIZE-CHARS";
		allTokens[SIZEPIXELS].keywordText = "SIZE-PIXELS";
		allTokens[SKIP].keywordText = "SKIP";
		allTokens[SKIPDELETEDRECORD].keywordText = "SKIP-DELETED-RECORD";
		allTokens[SLIDER].keywordText = "SLIDER";
		allTokens[SMALLINT].keywordText = "SMALLINT";
		allTokens[SOCKET].keywordText = "SOCKET";
		allTokens[SOME].keywordText = "SOME";
		allTokens[SORT].keywordText = "SORT";
		allTokens[SOURCE].keywordText = "SOURCE";
		allTokens[SOURCEPROCEDURE].keywordText = "SOURCE-PROCEDURE";
		allTokens[SPACE].keywordText = "SPACE";
		allTokens[SQL].keywordText = "SQL";
		allTokens[SQRT].keywordText = "SQRT";
		allTokens[START].keywordText = "START";
		allTokens[STARTMOVE].keywordText = "START-MOVE";
		allTokens[STARTRESIZE].keywordText = "START-RESIZE";
		allTokens[STARTROWRESIZE].keywordText = "START-ROW-RESIZE";
		allTokens[STATUS].keywordText = "STATUS";
		allTokens[STATUSBAR].keywordText = "STATUS-BAR";
		allTokens[STDCALL_KW].keywordText = "STDCALL";
		allTokens[STRETCHTOFIT].keywordText = "STRETCH-TO-FIT";
		allTokens[STOP].keywordText = "STOP";
		allTokens[STOREDPROCEDURE].keywordText = "STORED-PROCEDURE";
		allTokens[STREAM].keywordText = "STREAM";
		allTokens[STREAMIO].keywordText = "STREAM-IO";
		allTokens[STRING].keywordText = "STRING";
		allTokens[STRINGXREF].keywordText = "STRING-XREF";
		allTokens[SUBAVERAGE].keywordText = "SUB-AVERAGE";
		allTokens[SUBCOUNT].keywordText = "SUB-COUNT";
		allTokens[SUBMAXIMUM].keywordText = "SUB-MAXIMUM";
		allTokens[SUBMENU].keywordText = "SUB-MENU";
		allTokens[SUBMENUHELP].keywordText = "SUB-MENU-HELP";
		allTokens[SUBMINIMUM].keywordText = "SUB-MINIMUM";
		allTokens[SUBTOTAL].keywordText = "SUB-TOTAL";
		allTokens[SUBSCRIBE].keywordText = "SUBSCRIBE";
		allTokens[SUBSTITUTE].keywordText = "SUBSTITUTE";
		allTokens[SUBSTRING].keywordText = "SUBSTRING";
		allTokens[SUM].keywordText = "SUM";
		allTokens[SUMMARY].keywordText = "SUMMARY";
		allTokens[SUPER].keywordText = "SUPER";
		allTokens[SYSTEMDIALOG].keywordText = "SYSTEM-DIALOG";
		allTokens[SYSTEMHELP].keywordText = "SYSTEM-HELP";
		allTokens[TABLE].keywordText = "TABLE";
		allTokens[TABLEHANDLE].keywordText = "TABLE-HANDLE";
		allTokens[TABLENUMBER].keywordText = "TABLE-NUMBER";
		allTokens[TARGET].keywordText = "TARGET";
		allTokens[TARGETPROCEDURE].keywordText = "TARGET-PROCEDURE";
		allTokens[TEMPTABLE].keywordText = "TEMP-TABLE";
		allTokens[TERMINAL].keywordText = "TERMINAL";
		allTokens[TERMINATE].keywordText = "TERMINATE";
		allTokens[TEXT].keywordText = "TEXT";
		allTokens[TEXTCURSOR].keywordText = "TEXT-CURSOR";
		allTokens[TEXTSEGGROW].keywordText = "TEXT-SEG-GROWTH";
		allTokens[THEN].keywordText = "THEN";
		allTokens[THISPROCEDURE].keywordText = "THIS-PROCEDURE";
		allTokens[THREED].keywordText = "THREE-D";
		allTokens[THROUGH].keywordText = "THROUGH";
		allTokens[TICMARKS].keywordText = "TIC-MARKS";
		allTokens[TIME].keywordText = "TIME";
		allTokens[TITLE].keywordText = "TITLE";
		allTokens[TO].keywordText = "TO";
		allTokens[TOOLBAR].keywordText = "TOOL-BAR";
		allTokens[TOOLTIP].keywordText = "TOOLTIP";
		allTokens[TOROWID].keywordText = "TO-ROWID";
		allTokens[TODAY].keywordText = "TODAY";
		allTokens[TOGGLEBOX].keywordText = "TOGGLE-BOX";
		allTokens[TOP].keywordText = "TOP";
		allTokens[TOPONLY].keywordText = "TOP-ONLY";
		allTokens[TOPIC].keywordText = "TOPIC";
		allTokens[TOTAL].keywordText = "TOTAL";
		allTokens[TRANSACTION].keywordText = "TRANSACTION";
		allTokens[TRANSACTIONMODE].keywordText = "TRANSACTION-MODE";
		allTokens[TRANSPARENT].keywordText = "TRANSPARENT";
		allTokens[TRAILING].keywordText = "TRAILING";
		allTokens[TRIGGER].keywordText = "TRIGGER";
		allTokens[TRIGGERS].keywordText = "TRIGGERS";
		allTokens[TRIM].keywordText = "TRIM";
		allTokens[TRUE_KW].keywordText = "TRUE";
		allTokens[TRUNCATE].keywordText = "TRUNCATE";
		allTokens[UNBUFFERED].keywordText = "UNBUFFERED";
		allTokens[UNDERLINE].keywordText = "UNDERLINE";
		allTokens[UNDO].keywordText = "UNDO";
		allTokens[UNFORMATTED].keywordText = "UNFORMATTED";
		allTokens[UNION].keywordText = "UNION";
		allTokens[UNIQUE].keywordText = "UNIQUE";
		allTokens[UNIQUEMATCH].keywordText = "UNIQUE-MATCH";
		allTokens[UNIX].keywordText = "UNIX";
		allTokens[UNLESSHIDDEN].keywordText = "UNLESS-HIDDEN";
		allTokens[UNLOAD].keywordText = "UNLOAD";
		allTokens[UNSUBSCRIBE].keywordText = "UNSUBSCRIBE";
		allTokens[UP].keywordText = "UP";
		allTokens[UPDATE].keywordText = "UPDATE";
		allTokens[URLDECODE].keywordText = "URL-DECODE";
		allTokens[URLENCODE].keywordText = "URL-ENCODE";
		allTokens[USE].keywordText = "USE";
		allTokens[USEDICTEXPS].keywordText = "USE-DICT-EXPS";
		allTokens[USEFILENAME].keywordText = "USE-FILENAME";
		allTokens[USEINDEX].keywordText = "USE-INDEX";
		allTokens[USER].keywordText = "USER";
		allTokens[USERID].keywordText = "USERID";
		allTokens[USEREVVIDEO].keywordText = "USE-REVVIDEO";
		allTokens[USETEXT].keywordText = "USE-TEXT";
		allTokens[USEUNDERLINE].keywordText = "USE-UNDERLINE";
		allTokens[USING].keywordText = "USING";
		allTokens[V6FRAME].keywordText = "V6FRAME";
		allTokens[VALIDEVENT].keywordText = "VALID-EVENT";
		allTokens[VALIDHANDLE].keywordText = "VALID-HANDLE";
		allTokens[VALIDATE].keywordText = "VALIDATE";
		allTokens[VALUE].keywordText = "VALUE";
		allTokens[VALUECHANGED].keywordText = "VALUE-CHANGED";
		allTokens[VALUES].keywordText = "VALUES";
		allTokens[VARIABLE].keywordText = "VARIABLE";
		allTokens[VERBOSE].keywordText = "VERBOSE";
		allTokens[VERTICAL].keywordText = "VERTICAL";
		allTokens[VIEW].keywordText = "VIEW";
		allTokens[VIEWAS].keywordText = "VIEW-AS";
		allTokens[VMS].keywordText = "VMS";
		allTokens[WAIT].keywordText = "WAIT";
		allTokens[WAITFOR].keywordText = "WAIT-FOR";
		allTokens[WARNING].keywordText = "WARNING";
		allTokens[WEBCONTEXT].keywordText = "WEB-CONTEXT";
		allTokens[WEEKDAY].keywordText = "WEEKDAY";
		allTokens[WHEN].keywordText = "WHEN";
		allTokens[WHERE].keywordText = "WHERE";
		allTokens[WHILE].keywordText = "WHILE";
		allTokens[WIDGET].keywordText = "WIDGET";
		allTokens[WIDGETHANDLE].keywordText = "WIDGET-HANDLE";
		allTokens[WIDGETPOOL].keywordText = "WIDGET-POOL";
		allTokens[WIDTH].keywordText = "WIDTH";
		allTokens[WIDTHCHARS].keywordText = "WIDTH-CHARS";
		allTokens[WIDTHPIXELS].keywordText = "WIDTH-PIXELS";
		allTokens[WINDOW].keywordText = "WINDOW";
		allTokens[WINDOWDELAYEDMINIMIZE].keywordText = "WINDOW-DELAYED-MINIMIZE";
		allTokens[WINDOWMAXIMIZED].keywordText = "WINDOW-MAXIMIZED";
		allTokens[WINDOWMINIMIZED].keywordText = "WINDOW-MINIMIZED";
		allTokens[WINDOWNAME].keywordText = "WINDOW-NAME";
		allTokens[WINDOWNORMAL].keywordText = "WINDOW-NORMAL";
		allTokens[WITH].keywordText = "WITH";
		allTokens[WORDINDEX].keywordText = "WORD-INDEX";
		allTokens[WORKTABLE].keywordText = "WORK-TABLE";
		allTokens[WRITE].keywordText = "WRITE";
		allTokens[X].keywordText = "X";
		allTokens[XDOCUMENT].keywordText = "X-DOCUMENT";
		allTokens[XNODEREF].keywordText = "X-NODEREF";
		allTokens[XOF].keywordText = "X-OF";
		allTokens[XCODE].keywordText = "XCODE";
		allTokens[XREF].keywordText = "XREF";
		allTokens[Y].keywordText = "Y";
		allTokens[YOF].keywordText = "Y-OF";
		allTokens[YEAR].keywordText = "YEAR";
		allTokens[YES].keywordText = "YES";
		allTokens[YESNO].keywordText = "YES-NO";
		allTokens[YESNOCANCEL].keywordText = "YES-NO-CANCEL";
		allTokens[FUNCTIONCALLTYPE].keywordText = "FUNCTION-CALL-TYPE";
		allTokens[GETATTRCALLTYPE].keywordText = "GET-ATTR-CALL-TYPE";
		allTokens[PROCEDURECALLTYPE].keywordText = "PROCEDURE-CALL-TYPE";
		allTokens[SAXCOMPLETE].keywordText = "SAX-COMPLETE";
		allTokens[SAXPARSERERROR].keywordText = "SAX-PARSER-ERROR";
		allTokens[SAXRUNNING].keywordText = "SAX-RUNNING";
		allTokens[SAXUNINITIALIZED].keywordText = "SAX-UNINITIALIZED";
		allTokens[SETATTRCALLTYPE].keywordText = "SET-ATTR-CALL-TYPE";
		allTokens[COPYLOB].keywordText = "COPY-LOB";
		allTokens[DATARELATION].keywordText = "DATA-RELATION";
		allTokens[DATASOURCE].keywordText = "DATA-SOURCE";
		allTokens[DATASET].keywordText = "DATASET";
		allTokens[DATASETHANDLE].keywordText = "DATASET-HANDLE";
		allTokens[LOGMANAGER].keywordText = "LOG-MANAGER";
		allTokens[NOLOBS].keywordText = "NO-LOBS";
		allTokens[NOW].keywordText = "NOW";
		allTokens[STARTING].keywordText = "STARTING";
		allTokens[BASE64].keywordText = "BASE64";
		allTokens[SOAPHEADER].keywordText = "SOAP-HEADER";
		allTokens[SOAPHEADERENTRYREF].keywordText = "SOAP-HEADER-ENTRYREF";
		allTokens[BLOB].keywordText = "BLOB";
		allTokens[CLOB].keywordText = "CLOB";
		allTokens[DATETIME].keywordText = "DATETIME";
		allTokens[DATETIMETZ].keywordText = "DATETIME-TZ";
		allTokens[LONGCHAR].keywordText = "LONGCHAR";
		allTokens[RELATIONFIELDS].keywordText = "RELATION-FIELDS";
		allTokens[TTCODEPAGE].keywordText = "TTCODEPAGE";
		allTokens[COLUMNCODEPAGE].keywordText = "COLUMN-CODEPAGE";
		allTokens[DYNAMICCURRENTVALUE].keywordText = "DYNAMIC-CURRENT-VALUE";
		allTokens[DYNAMICNEXTVALUE].keywordText = "DYNAMIC-NEXT-VALUE";
		allTokens[FIXCODEPAGE].keywordText = "FIX-CODEPAGE";
		allTokens[INTERVAL].keywordText = "INTERVAL";
		allTokens[ISCODEPAGEFIXED].keywordText = "IS-CODEPAGE-FIXED";
		allTokens[ISCOLUMNCODEPAGE].keywordText = "IS-COLUMN-CODEPAGE";
		allTokens[ISODATE].keywordText = "ISO-DATE";
		allTokens[MTIME].keywordText = "MTIME";
		allTokens[LOBDIR].keywordText = "LOB-DIR";
		allTokens[TIMEZONE].keywordText = "TIMEZONE";
		allTokens[BYVALUE].keywordText = "BY-VALUE";
		allTokens[BYREFERENCE].keywordText = "BY-REFERENCE";
		allTokens[ADDINTERVAL].keywordText = "ADD-INTERVAL";
		allTokens[GETDIR].keywordText = "GET-DIR";
		allTokens[CURRENCY].keywordText = "CURRENCY";
		allTokens[ERRORCODE].keywordText = "ERROR-CODE";
		allTokens[IUNKNOWN].keywordText = "IUNKNOWN";
		allTokens[SHORT].keywordText = "SHORT";
		allTokens[UNSIGNEDBYTE].keywordText = "UNSIGNED-BYTE";
		allTokens[UNSIGNEDSHORT].keywordText = "UNSIGNED-SHORT";
		allTokens[CODEPAGE].keywordText = "CODEPAGE";
		allTokens[BASE64DECODE].keywordText = "BASE64-DECODE";
		allTokens[BASE64ENCODE].keywordText = "BASE64-ENCODE";
		allTokens[BATCHSIZE].keywordText = "BATCH-SIZE";
		allTokens[BEFORETABLE].keywordText = "BEFORE-TABLE";
		allTokens[COPYDATASET].keywordText = "COPY-DATASET";
		allTokens[COPYTEMPTABLE].keywordText = "COPY-TEMP-TABLE";
		allTokens[DATASOURCEMODIFIED].keywordText = "DATA-SOURCE-MODIFIED";
		allTokens[DECRYPT].keywordText = "DECRYPT";
		allTokens[DELETECHARACTER].keywordText = "DELETE-CHARACTER";
		allTokens[ENABLEDFIELDS].keywordText = "ENABLED-FIELDS";
		allTokens[ENCRYPT].keywordText = "ENCRYPT";
		allTokens[ENCRYPTIONSALT].keywordText = "ENCRYPTION-SALT";
		allTokens[FORMLONGINPUT].keywordText = "FORM-LONG-INPUT";
		allTokens[GENERATEPBEKEY].keywordText = "GENERATE-PBE-KEY";
		allTokens[GENERATEPBESALT].keywordText = "GENERATE-PBE-SALT";
		allTokens[GENERATERANDOMKEY].keywordText = "GENERATE-RANDOM-KEY";
		allTokens[GETCGILONGVALUE].keywordText = "GET-CGI-LONG-VALUE";
		allTokens[LASTBATCH].keywordText = "LAST-BATCH";
		allTokens[MD5DIGEST].keywordText = "MD5-DIGEST";
		allTokens[MERGEBYFIELD].keywordText = "MERGE-BY-FIELD";
		allTokens[NORMALIZE].keywordText = "NORMALIZE";
		allTokens[PBEHASHALGORITHM].keywordText = "PBE-HASH-ALGORITHM";
		allTokens[PBEKEYROUNDS].keywordText = "PBE-KEY-ROUNDS";
		allTokens[PREFERDATASET].keywordText = "PREFER-DATASET";
		allTokens[REJECTED].keywordText = "REJECTED";
		allTokens[REPOSITIONMODE].keywordText = "REPOSITION-MODE";
		allTokens[ROWSTATE].keywordText = "ROW-STATE";
		allTokens[ROWUNMODIFIED].keywordText = "ROW-UNMODIFIED";
		allTokens[ROWDELETED].keywordText = "ROW-DELETED";
		allTokens[ROWMODIFIED].keywordText = "ROW-MODIFIED";
		allTokens[ROWCREATED].keywordText = "ROW-CREATED";
		allTokens[SECURITYPOLICY].keywordText = "SECURITY-POLICY";
		allTokens[SHA1DIGEST].keywordText = "SHA1-DIGEST";
		allTokens[SSLSERVERNAME].keywordText = "SSL-SERVER-NAME";
		allTokens[SYMMETRICENCRYPTIONALGORITHM].keywordText = "SYMMETRIC-ENCRYPTION-ALGORITHM";
		allTokens[SYMMETRICENCRYPTIONIV].keywordText = "SYMMETRIC-ENCRYPTION-IV";
		allTokens[SYMMETRICENCRYPTIONKEY].keywordText = "SYMMETRIC-ENCRYPTION-KEY";
		allTokens[SYMMETRICSUPPORT].keywordText = "SYMMETRIC-SUPPORT";
		allTokens[TRANSINITPROCEDURE].keywordText = "TRANS-INIT-PROCEDURE";

		// OpenEdge 10.1, Proparse 3.0.
		allTokens[BIGINT].keywordText = "BIGINT";
		allTokens[TIMESTAMP].keywordText = "TIMESTAMP";
		allTokens[FIXCHAR].keywordText = "FIXCHAR";
		allTokens[AUDITCONTROL].keywordText = "AUDIT-CONTROL";
		allTokens[AUDITENABLED].keywordText = "AUDIT-ENABLED";
		allTokens[AUDITPOLICY].keywordText = "AUDIT-POLICY";
		allTokens[BIND].keywordText = "BIND";
		allTokens[CAST].keywordText = "CAST";
		allTokens[CLASS].keywordText = "CLASS";
		allTokens[CLIENTPRINCIPAL].keywordText = "CLIENT-PRINCIPAL";
		allTokens[CONSTRUCTOR].keywordText = "CONSTRUCTOR";
		allTokens[FINAL].keywordText = "FINAL";
		allTokens[GENERATEUUID].keywordText = "GENERATE-UUID";
		allTokens[GUID].keywordText = "GUID";
		allTokens[HEXDECODE].keywordText = "HEX-DECODE";
		allTokens[HEXENCODE].keywordText = "HEX-ENCODE";
		allTokens[IMPLEMENTS].keywordText = "IMPLEMENTS";
		allTokens[INHERITS].keywordText = "INHERITS";
		allTokens[INTERFACE].keywordText = "INTERFACE";
		allTokens[METHOD].keywordText = "METHOD";
		allTokens[NAMESPACEPREFIX].keywordText = "NAMESPACE-PREFIX";
		allTokens[NAMESPACEURI].keywordText = "NAMESPACE-URI";
		allTokens[NEWINSTANCE].keywordText = "NEW-INSTANCE";
		allTokens[PROTECTED].keywordText = "PROTECTED";
		allTokens[REFERENCEONLY].keywordText = "REFERENCE-ONLY";
		allTokens[SAXWRITER].keywordText = "SAX-WRITER";
		allTokens[SETDBCLIENT].keywordText = "SET-DB-CLIENT";
		allTokens[THISOBJECT].keywordText = "THIS-OBJECT";
		allTokens[TYPEOF].keywordText = "TYPE-OF";
		allTokens[VALIDOBJECT].keywordText = "VALID-OBJECT";
		allTokens[XMLDATATYPE].keywordText = "XML-DATA-TYPE";
		allTokens[XMLNODETYPE].keywordText = "XML-NODE-TYPE";
		allTokens[WIDGETID].keywordText = "WIDGET-ID";
		allTokens[DESTRUCTOR].keywordText = "DESTRUCTOR";
		allTokens[VOID].keywordText = "VOID";
		allTokens[NESTED].keywordText = "NESTED";
		allTokens[ROUNDED].keywordText = "ROUNDED";
		allTokens[GROUPBOX].keywordText = "GROUP-BOX";
		allTokens[PASSWORDFIELD].keywordText = "PASSWORD-FIELD";
		
		// OpenEdge 10.1B, Proparse 3.1
		allTokens[ROUNDED].keywordText = "ROUNDED";
		allTokens[GROUPBOX].keywordText = "GROUP-BOX";
		allTokens[FALSELEAKS].keywordText = "FALSE-LEAKS";
		allTokens[LEAKDETECTION].keywordText = "LEAK-DETECTION";
		allTokens[SAXWRITEBEGIN].keywordText = "SAX-WRITE-BEGIN";
		allTokens[SAXWRITECOMPLETE].keywordText = "SAX-WRITE-COMPLETE";
		allTokens[SAXWRITECONTENT].keywordText = "SAX-WRITE-CONTENT";
		allTokens[SAXWRITEELEMENT].keywordText = "SAX-WRITE-ELEMENT";
		allTokens[SAXWRITEERROR].keywordText = "SAX-WRITE-ERROR";
		allTokens[SAXWRITEIDLE].keywordText = "SAX-WRITE-IDLE";
		allTokens[SAXWRITETAG].keywordText = "SAX-WRITE-TAG";
		allTokens[STOMPDETECTION].keywordText = "STOMP-DETECTION";
		allTokens[STOMPFREQUENCY].keywordText = "STOMP-FREQUENCY";
		allTokens[INT64].keywordText = "INT64";
		allTokens[PUTINT64].keywordText = "PUT-INT64";
		allTokens[GETINT64].keywordText = "GET-INT64";
		allTokens[PUTUNSIGNEDLONG].keywordText = "PUT-UNSIGNED-LONG";
		allTokens[GETUNSIGNEDLONG].keywordText = "GET-UNSIGNED-LONG";
		allTokens[PROPERTY].keywordText = "PROPERTY";
		allTokens[SAXATTRIBUTES].keywordText = "SAX-ATTRIBUTES";
		allTokens[INHERITBGCOLOR].keywordText = "INHERIT-BGCOLOR";
		allTokens[NOINHERITBGCOLOR].keywordText = "NO-INHERIT-BGCOLOR";
		allTokens[INHERITFGCOLOR].keywordText = "INHERIT-FGCOLOR";
		allTokens[NOINHERITFGCOLOR].keywordText = "NO-INHERIT-FGCOLOR";
		allTokens[USEWIDGETPOOL].keywordText = "USE-WIDGET-POOL";
		allTokens[XREFXML].keywordText = "XREF-XML";

		// 10.1C and 10.2 keywords.
		allTokens[ACTIVEFORM].keywordText = "ACTIVE-FORM";
		allTokens[ASSEMBLY].keywordText = "ASSEMBLY";
		allTokens[CATCH].keywordText = "CATCH";
		allTokens[CREATELIKESEQUENTIAL].keywordText = "CREATE-LIKE-SEQUENTIAL";
		allTokens[CURRENTQUERY].keywordText = "CURRENT-QUERY";
		allTokens[DATASOURCEROWID].keywordText = "DATA-SOURCE-ROWID";
		allTokens[DEFAULTVALUE].keywordText = "DEFAULT-VALUE";
		allTokens[ERRORSTACKTRACE].keywordText = "ERROR-STACK-TRACE";
		allTokens[FINALLY].keywordText = "FINALLY";
		allTokens[FIRSTFORM].keywordText = "FIRST-FORM";
		allTokens[LASTFORM].keywordText = "LAST-FORM";
		allTokens[LIKESEQUENTIAL].keywordText = "LIKE-SEQUENTIAL";
		allTokens[MARKNEW].keywordText = "MARK-NEW";
		allTokens[MARKROWSTATE].keywordText = "MARK-ROW-STATE";
		allTokens[MAXIMUMLEVEL].keywordText = "MAXIMUM-LEVEL";
		allTokens[NOTACTIVE].keywordText = "NOT-ACTIVE";
		allTokens[RESTARTROW].keywordText = "RESTART-ROW";
		allTokens[ROUTINELEVEL].keywordText = "ROUTINE-LEVEL";
		allTokens[STATIC].keywordText = "STATIC";
		allTokens[STREAMHANDLE].keywordText = "STREAM-HANDLE";
		allTokens[THROW].keywordText = "THROW";
		allTokens[TOPNAVQUERY].keywordText = "TOP-NAV-QUERY";
		allTokens[UNBOX].keywordText = "UNBOX";
		allTokens[BOX].keywordText = "BOX";
		allTokens[DBREMOTEHOST].keywordText = "DB-REMOTE-HOST";
		allTokens[DYNAMICCAST].keywordText = "DYNAMIC-CAST";
		allTokens[XMLNODENAME].keywordText = "XML-NODE-NAME";
		allTokens[FOREIGNKEYHIDDEN].keywordText = "FOREIGN-KEY-HIDDEN";

	}
}
